# Provide version-number etc.
# This file is automatically edited by git!
# This file must not have any sideeffects or do anything but define meta variables!
from ._version import get_versions

__version__ = get_versions()['version']
__copyright__ = '2023, Fraunhofer IEE'
__title__ = 'CurriculumAgent'
__description__ = 'CurriculumAgent is a cleanup and improved version of the NeurIPS 2020 Competition Agent by ' \
                  'Binbinchen.The agent is build to extract action sets of the Grid2Op Environment and then ' \
                  'use rule-based agent to train a Reinforcement Learning agent.'
__author__ = 'OE224'
__authors__ = ['Malte Lehna']
